from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from types import TracebackType


class Player(AbstractContextManager, ABC):
    """
    Class for wrapping different AIs.
    """

    def __init__(self, logs: list[str] | None = None) -> None:
        self._logs: list[str] = logs or []

    def __enter__(self) -> "Player":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        pass

    @abstractmethod
    def play(self, move: str) -> str:
        """
        Play a move.

        Args:
            move (str): Previous move

        Returns:
            str: New move
        """

    def pop_logs(self) -> str:
        """
        Return and clear all logs in memory

        Returns:
            str: Logs
        """

        logs = "\n".join(self._logs)
        self._logs = []

        return logs



from enum import Enum


class GameState(Enum):
    CONTINUE = "CONTINUE"
    WIN = "WIN"
    LOSE = "LOSE"
    DRAW = "DRAW"
    INVALID = "INVALID"
    ILLEGAL = "ILLEGAL"
    MAX_TURNS = "MAX_TURNS"
    TIMEOUT = "TIMEOUT"

    def __str__(self):
        return self.value



import time
from collections.abc import Callable
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Any

from duo_game_lib.game_state import GameState
from duo_game_lib.judge import Judge
from duo_game_lib.move import Move, MoveMetadata
from duo_game_lib.player import Player


class Game(AbstractContextManager):
    def __init__(
        self,
        logger: Callable[[str], None],
        player1: Player,
        player2: Player,
        judge: Judge,
    ) -> None:
        self.__logger: Callable[[Any], None] = logger
        self.__players: list[Player] = [player1, player2]
        self.__judge: Judge = judge

    def __enter__(self) -> Any:
        for player in self.__players:
            player.__enter__()

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        for player in self.__players:
            player.__exit__(exc_type, exc_value, traceback)

    def play(self, turns: int = 250, debug: bool = False):
        """
        Starts a game and return result as dict.

        Args:
            turns (int, optional): The maximum amount of turns to play. Defaults to 250.
            debug(bool, optional): If True, prints debug info to console. Defaults to False.

        Returns:
            dict[str, Any]: The game result containing winner, moves, etc.
        """
        previous_move = ""
        state = GameState.CONTINUE
        move = None

        i = 0

        while state == GameState.CONTINUE:
            if i == turns:
                state = GameState.MAX_TURNS
                break

            player = self.__players[i % 2]
            move, elapsed_time = self.__play_one_move(player, previous_move)

            if elapsed_time >= 60000:
                state = GameState.TIMEOUT
                continue

            state = self.__judge.validate(move)

            if state == GameState.CONTINUE:
                self.__judge.add_move(move)
                state = self.__judge.is_game_over()

            move_object = Move(
                move,
                state,
                MoveMetadata(
                    elapsed_time,
                    self.__judge.analyze(),
                    player.get_and_reset_current_logs(),
                ),
            )
            self.__logger(f"MOVE: {move_object}")

            previous_move = move

            if debug:
                self.__print_debug_info(move_object)

            i += 1

        self.__logger(f"END: {state}")

    def __play_one_move(self, player: Player, prev_move: str) -> tuple[str, int]:
        start_time = time.perf_counter()

        move = player.play(prev_move)

        elapsed_time = int((time.perf_counter() - start_time) * 1000)

        return (move, elapsed_time)

    def __print_debug_info(self, move: Move) -> None:
        info = "\n".join(
            [
                self.__judge.get_debug_info(),
                str(move),
            ]
        )

        print(info)



from abc import ABC, abstractmethod
from typing import Any

from duo_game_lib.game_state import GameState


class Judge(ABC):
    """
    An abstract class for validating moves and generating debug infos of a game.
    """

    @abstractmethod
    def is_game_over(self) -> GameState:
        """
        Returns corresponding ending state if game over, otherwise CONTINUE.

        Returns:
            GameState: State of game.
        """

    @abstractmethod
    def validate(self, move: str) -> GameState:
        """
        Validates a move, does not change internal states.

        Args:
            move (str): Move to be validated.

        Returns:
            GameState: An integer describing game state.
        """

    @abstractmethod
    def add_move(self, move: str) -> Any:
        """
        Adds a move and updates internal data structures based on game.

        Args:
            move (str): The move to be added.

        Returns:
            Any: Additional information of added more when needed.
        """

    @abstractmethod
    def get_debug_info(self) -> str:
        """
        Returns debug info such as visual board state, player advantages, etc.

        Returns:
            str: Debug info.
        """

    @abstractmethod
    def get_all_moves(self) -> list[str]:
        """
        Returns all moves as string, formatted based on game.

        Returns:
            list[str]: List of moves.
        """

    @abstractmethod
    def analyze(self) -> float | int:
        """
        Evaluation game situation.
        Depends on game and may need additional arguments to evaluate situation.

        Returns:
            float | int: Evaluation of game.
        """



import json

from duo_game_lib.game_state import GameState


class MoveMetadata:
    def __init__(self, time: int, evaluation: float | int, logs: str):
        self.time: int = time
        self.evaluation: float | int = evaluation
        self.logs: str = logs


class Move:
    def __init__(self, move: str, state: GameState, metadata: MoveMetadata) -> None:
        self.move: str = move
        self.state: GameState = state
        self.time: int = metadata.time
        self.evaluation: float | int = metadata.evaluation
        self.logs: str = metadata.logs

    def __str__(self) -> str:
        result = {
            "move": self.move,
            "state": self.state.name,
            "time": self.time,
            "evaluation": self.evaluation,
            "logs": self.logs,
        }

        return json.dumps(result)

    def as_json(self):
        return {
            "move": self.move,
            "state": str(self.state),
            "time": self.time,
            "evaluation": self.evaluation,
            "logs": self.logs,
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Move):
            return False

        return (
            self.move == other.move
            and self.state == other.state
            and self.time == other.time
            and self.evaluation == other.evaluation
            and self.logs == other.logs
        )
